// MO6export
// by Michele Alessandrini, drinka@libero.it
// You can do with this source whatever you want, but a citation to the author would be appreciated.
// Read carefully comments about constants defined below.
// If you compile with DUMP_DATA defined, a modified image for every frame will be saved, with grid
// points marked in white, for debug purpose.

#include <iostream>
#include <string>
#include <utility>
using namespace std;
#include <stdio.h>

// IMPORTANT: set these constants to your captured images size. I suggest to capture at least at
// 640 * 480, I didn't test it with smaller resolutions.
const unsigned width=640, height=480;

// IMPORTANT: you must say to the program where are the centers of the upper left green square and
// of the lower right one, the grid will be calculated automatically from these values. Capture a
// sample image, then open it with "paint" or other minimal program, annotate the centers of such
// squares and write it in p1x, p1y, p2x, p2y. Try to be accurate in finding centers.
const unsigned p1x=95, p1y=73, p2x=586, p2y=398;

// END OF REQUIRED ADJUSTMENTS

typedef unsigned char UC;
const UC threshold = 128;

bool isRed(UC *bgr) {return bgr[0]<threshold && bgr[1]<threshold && bgr[2]>UC(255)-threshold;}
bool isGreen(UC *bgr) {return bgr[0]<threshold && bgr[1]>UC(255)-threshold && bgr[2]<threshold;}
bool isBlack(UC *bgr) {return bgr[0]<threshold && bgr[1]<threshold && bgr[2]<threshold;}

struct BMPHeader {
	//put 2 chars 'BM' before header
	unsigned fileSize, reserved, dataOffset, headerSize, width, height;
	unsigned short planes, bitsPerPixel;
	unsigned compression, dataSize, hRes /*0x400*/, vRes /*0x313*/, colors, impColors;
	BMPHeader(unsigned w, unsigned h, unsigned short bpp, unsigned dataSize):
		fileSize(dataSize+54),reserved(0),dataOffset(54),headerSize(0x28),width(w),height(h),planes(1),bitsPerPixel(bpp),
		compression(0),dataSize(dataSize),hRes(0xec4),vRes(0xec4),colors(0),impColors(0) {}
};

bool writeBMP(const char *fileName, unsigned width, unsigned height, unsigned short bpp, void *data, unsigned dataSize, unsigned repeat=1) {
	FILE *fout = fopen(fileName,"wb");
	if (!fout) return false;
	char *id = "BM";
	if (fwrite(id,1,2,fout)!=2) {
		fclose(fout);
		return false;
	}
	BMPHeader header(width,height,bpp,dataSize*repeat);
	if (fwrite(&header,sizeof(header),1,fout)!=1) {
		fclose(fout);
		return false;
	}
	for (unsigned i=0; i<repeat; ++i)
		if (fwrite(data,1,dataSize,fout)!=dataSize) {
			fclose(fout);
			return false;
		}
	fclose(fout);
	return true;	
}

bool readBMP(const char *fileName, unsigned width, unsigned height, unsigned short bpp, void *data, unsigned *bytesRead) {
	FILE *fin = fopen(fileName,"rb");
	if (!fin) return false;
	char id[2];
	if (fread(id,1,2,fin)!=2 || id[0]!='B' || id[1]!='M') {
		fclose(fin);
		return false;
	}
	BMPHeader h(0,0,0,0);
	if (fread(&h,sizeof(h),1,fin)!=1) {
		fclose(fin);
		return false;
	}
	if (h.fileSize!=h.dataSize+54 || h.reserved || h.dataOffset!=54 || h.headerSize!=0x28 || h.width!=width || h.height!=height
				|| h.planes!=1 || h.bitsPerPixel!=bpp || h.compression || h.colors || h.impColors) {
		fclose(fin);
		return false;
	}
	if (fread(data,1,h.dataSize,fin)!=h.dataSize) {
		fclose(fin);
		return false;
	}
	fclose(fin);
	*bytesRead=h.dataSize;
	return true;	
}

bool readFrame(const string &fileNamePrefix, UC numDigits, UC frameNum, UC *buffer) {  //frameNum 1-based, buffer size == 120
	UC data[width*height*3], *ptr;
	unsigned bytesRead;
	string fileName(fileNamePrefix), fileNum;
	UC f=frameNum;
	while (f) {
		fileNum=char('0'+f%10)+fileNum;
		f/=10;
	}
	while (fileNum.size()<unsigned(numDigits)) fileNum='0'+fileNum;
	fileName+=fileNum+".bmp";
	if (!readBMP(fileName.c_str(),width,height,24,data,&bytesRead) || bytesRead!=width*height*3) {
		cerr << "Error reading file " << fileName << endl;
		return false;
	}
#ifdef DUMP_DATA
	cout << "# " << fileName << endl;
#endif	
	double stepX = double(p2x-p1x)/39., stepY = double(p2y-p1y)/24.;
	pair<unsigned,unsigned> points[1000], *cur=points;
	for (unsigned r=0; r<25; ++r)
		for (unsigned c=0; c<40; ++c)
			*cur++ = pair<unsigned,unsigned>(p1x+c*stepX,p1y+r*stepY);

	cur=points;
	for (unsigned i=0; i<120; buffer[i++]=0);
	UC bits[] = {128,64,32,16,8,4,2,1};
	for (unsigned i=0; i<120; ++i) 
		for (unsigned b=0; b<8; ++b) {
			ptr=data + 3*width*(height-cur->second-1) + 3*cur->first;
			if (isGreen(ptr)) buffer[i]+=bits[b];
			else if (!isBlack(ptr)) {cerr << "Invalid color in pattern in file " << fileName << endl; return false;}
#ifdef DUMP_DATA
			//modify data for debug
			ptr[0]=ptr[1]=ptr[2]=255;
#endif
			++cur;
		}
	UC control[4] = {0,0,0,0};
	for (unsigned i=0; i<4; ++i) 
		for (unsigned b=0; b<8; ++b) {
			ptr=data + 3*width*(height-cur->second-1) + 3*cur->first;
			if (isGreen(ptr)) control[i]+=bits[b];
			else if (!isBlack(ptr)) {cerr << "Invalid color in pattern in file " << fileName << endl; return false;}
			++cur;
		}
#ifdef DUMP_DATA
	if (!writeBMP((fileName+".mod.bmp").c_str(),width,height,24,data,width*height*3)) {cerr << "Error rewriting" << endl; return false;}
#endif
	unsigned short checksum=0;
	for (unsigned i=0; i<120; checksum+=buffer[i++]);
	if ((unsigned short)(control[0])*256+(unsigned short)(control[1])!=checksum || control[2]!=frameNum) {
		cerr << "Control sequence error in file " << fileName << " " << checksum << " " << unsigned(control[0]) << " "
			<< unsigned(control[1]) << " " << unsigned(control[2]) << endl;
		return false;
	}
	return true;
}

bool writeK7(const string &fileName, unsigned short loadAddr, unsigned short execAddr, const UC *buffer, unsigned dataSize) {
	UC *raw = new UC[dataSize+10], *rawp=raw+5;
	const UC *bufp=buffer;
	raw[0]=0; raw[1]=dataSize/256; raw[2]=dataSize%256; raw[3]=loadAddr/256; raw[4]=loadAddr%256;
	for (unsigned i=0; i<dataSize; ++i) *rawp++=*bufp++;
	*rawp++=255; *rawp++=0; *rawp++=0; *rawp++=execAddr/256; *rawp++=execAddr%256;
	UC header[18] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0x3c,0x5a};
	UC beginBlock[17] = {0,16,'B','I','N',32,32,32,32,32,'B','I','N',2,0,0,0xac};
	FILE *fout = fopen((fileName+".k7").c_str(),"wb");
	if (!fout || fwrite(header,18,1,fout)!=1 || fwrite(beginBlock,17,1,fout)!=1) {cerr << "Unable to write k7 file" << endl; return false;}
	unsigned rawSize=dataSize+10;
	rawp=raw;
	while (rawSize) {
		unsigned length = rawSize>=254 ? 256 : rawSize+2;
		UC chk=0, block[257], *b=block+2;
		block[0]=1; block[1]=length==256?0:length;
		for (unsigned i=0; i<length-2; ++i) {
			chk+=*rawp;
			*b++=*rawp++;
		}
		*b=256-chk;
		rawSize-=length-2;
		if (fwrite(header,18,1,fout)!=1 || fwrite(block,length+1,1,fout)!=1) {cerr << "Unable to write k7 file" << endl; return false;}
	}
	UC endBlock[3] = {255,2,0};
	if (fwrite(header,18,1,fout)!=1 || fwrite(endBlock,3,1,fout)!=1) {cerr << "Unable to write k7 file" << endl; return false;}
	fclose(fout);
	cout << "Wrote k7 file with parameters " << loadAddr << "," << loadAddr+dataSize-1 << "," << execAddr << endl;
	return true;
}

int main() {
	UC buffer[120*255], *cur=buffer;
	string prefixName, k7FileName;
	unsigned numDigits, framesNum;
	unsigned short loadAddr, execAddr;
	cout << "MO6export, by Michele Alessandrini, drinka@libero.it" << endl;
	cout << "Prefix name: "; cin >> prefixName;
	cout << "Number of digits in number: "; cin >> numDigits;
	cout << "Number of frames: "; cin >> framesNum;
	for (unsigned i=1; i<=framesNum; ++i,cur+=120)
		if (!readFrame(prefixName,numDigits,i,cur)) return 1;
	cout << "Recognition OK" << endl;
	cout << "K7 file name: "; cin >> k7FileName;
	cout << "Load address: "; cin >> loadAddr;
	cout << "Exec address: "; cin >> execAddr;
	if (!writeK7(k7FileName,loadAddr,execAddr,buffer,framesNum*120)) return 1;
	return 0;
};
